###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import json
import time
import numpy
import uuid
import mimetypes

import moviepy.editor as mp
from PIL import Image, ImageDraw, ImageFont
from googleapiclient.http import MediaFileUpload

from starthinker.util.project import project
from starthinker.util.sheets import sheets_read
from starthinker.util.bigquery import query_to_rows
from starthinker.util.storage import object_put
from starthinker.util.google_api import API_YouTube

from starthinker.config import BUFFER_SCALE
CHUNKSIZE = int(10 * 1024000 * BUFFER_SCALE)  # scale is controlled in config.py


def hex_to_rgb(value, alpha=None):
  # if string assume hex format ( supports both #ffffff and #fff notation )
  if isinstance(value, str):
    value = value.lstrip('#')
    hlen = len(value)
    rgba = [
        int(value[i:i + int(hlen / 3)], 16)
        for i in range(0, hlen, int(hlen / 3))
    ]
    if alpha:
      rgba.append(int(alpha * 255))
  # probably already a tuple just pass the value
  else:
    rgba = value
  return tuple(rgba)


def get_text_image(effect):

  # get a font ( from standard library )
  fnt = ImageFont.truetype('%s.ttf' % effect['text']['font'],
                           effect['text']['size'])

  # determine size of text
  img = Image.new('RGBA', (1, 1))
  draw = ImageDraw.Draw(img)
  size = draw.multiline_textsize(
      effect['text']['message'],
      font=fnt,
      spacing=int(effect['text']['size'] / 2))

  # make a blank image for the text and draw the text
  #print('S', size)
  txt = Image.new('RGBA', size, (255, 255, 255, 0))
  d = ImageDraw.Draw(txt)
  #print('S', effect['text']['color'])
  d.text((0, 0),
         effect['text']['message'],
         font=fnt,
         fill=hex_to_rgb(effect['text']['color'], effect.get('opacity')),
         align=effect['text']['align'])

  return numpy.array(txt)


def get_effects(video, effect):

  if 'image' in effect:
    clip = (mp.ImageClip(effect['image']['file_or_url']))

    clip = clip.set_start(effect['duration']['start'])
    clip = clip.set_duration(effect['duration']['end'] -
                             effect['duration']['start'])
    clip = clip.set_pos(
        (effect['position']['width'], effect['position']['height']))

    if effect.get('fade', {}).get('in'):
      clip = clip.crossfadein(effect['fade']['in'])

    if effect.get('fade', {}).get('out'):
      clip = clip.crossfadeout(effect['fade']['out'])

    if effect.get('position', {}).get('rotate'):
      clip = clip.rotate(effect['position']['rotate'])

    yield clip

  if 'text' in effect:

    # Requires working installation of ImageMagick
    try:
      clip = mp.TextClip(
          txt=effect['text']['message'],
          color=effect['text'].get('color', '#666666'),
          font=effect['text'].get('font', 'Courier'),
          fontsize=effect['text'].get('size', 12),
          align=effect['text'].get('align', 'center'),
          kerning=effect['text'].get('kerning', 0))

    # Alternate method using Pillow - no need to set position text is already positioned within image
    except:
      clip = (mp.ImageClip(get_text_image(effect)))

    clip = clip.set_start(effect['duration']['start'])
    clip = clip.set_duration(effect['duration']['end'] -
                             effect['duration']['start'])
    clip = clip.set_pos(
        (effect['position']['width'], effect['position']['height']))

    if effect.get('fade', {}).get('in'):
      clip = clip.crossfadein(effect['fade']['in'])

    if effect.get('fade', {}).get('out'):
      clip = clip.crossfadeout(effect['fade']['out'])

    if effect.get('position', {}).get('rotate'):
      clip = clip.rotate(effect['position']['rotate'])

    yield clip

  elif 'audio' in effect:
    clip = (mp.AudioClip(effect['audio']['file_or_url']))

    clip = clip.set_start(effect['duration']['start'])
    clip = clip.set_duration(effect['duration']['end'] -
                             effect['duration']['start'])

    yield clip


def edit_video(video):
  clips = [mp.VideoFileClip(video['file_or_url'])]

  for effect in video['effects']:
    clips.extend(get_effects(clips[0], effect))

  video = mp.CompositeVideoClip(clips)

  return video


def save_video(out, clip):
  # write to user defined local file
  if out.get('file'):
    clip.write_videofile(
        out['file'], codec='libx264', audio_codec='aac',
        logger=None)  # logger needed or fmmpeg writes to stderr

  # for storage, write to temporary file ( no alternative with moviepy ), then upload
  if out.get('storage', {}).get('file'):
    temporary_file_name = '/tmp/%s_%s' % (uuid.uuid1(), out['storage']['file'])
    clip.write_videofile(
        temporary_file_name, codec='libx264', audio_codec='aac',
        logger=None)  # logger needed or fmmpeg writes to stderr
    with open(temporary_file_name, 'rb') as temporary_file:
      object_put(
          project.task['auth'],
          '%s:%s' % (out['storage']['bucket'], out['storage']['file']),
          temporary_file,
          mimetype=mimetypes.guess_type(out['storage']['file'],
                                        strict=False)[0])
    os.remove(temporary_file_name)

  if out.get('dcm'):
    print('DCM not implemented yet.')

  # for youtube, write to temporary file ( no alternative with moviepy ), then upload
  if out.get('youtube', {}).get('title'):
    temporary_file_name = '/tmp/%s_%s' % (uuid.uuid1(), out['storage']['file'])
    clip.write_videofile(
        temporary_file_name, codec='libx264', audio_codec='aac',
        logger=None)  # logger needed or fmmpeg writes to stderr

    body = {
        'snippet': {
            'title': out['youtube']['title'],
            'description': out['youtube']['description'],
            'tags': out['youtube']['tags'],
            'categoryId': out['youtube']['category']
        },
        'status': {
            'privacyStatus': out['youtube']['privacy']
        }
    }
    try:
      API_YouTube(project.task['auth']).videos().insert(
          part=','.join(body.keys()),
          body=body,
          media_body=MediaFileUpload(
              temporary_file_name, chunksize=CHUNKSIZE,
              resumable=True)).upload()
    finally:
      os.remove(temporary_file_name)


def rows_to_videos(rows):
  videos = {}
  for row in rows:
    videos.setdefault(
        row[0], {
            'file_or_url': row[0],
            'effects': [],
            'out': {
                'youtube': {
                    'title': row[17],
                    'description': row[18],
                    'category': int(row[19]) if row[19] else None,
                    'tags': row[20].split(',') if row[20] else None,
                    'privacy': row[21].lower() if row[21] else None,
                },
                'dcm': row[22],
                'storage': {
                    'bucket': row[23],
                    'file': row[24],
                }
            }
        })['effects'].append({
            'duration': {
                'start': float(row[1]),
                'end': float(row[2])
            },
            'fade': {
                'in': float(row[3]),
                'out': float(row[4])
            },
            'opacity': float(row[5]),
            'position': {
                'height': int(row[6]),
                'width': int(row[7]),
                'angle': int(row[8])
            },
            'image': {
                'file_or_url': row[9]
            },
            'text': {
                'color': row[10],
                'font': row[11],
                'size': int(row[12]),
                'align': row[13],
                'kerning': int(row[14]),
                'message': row[15]
            }
        })
  return videos.values()


def videos_from_sheets(sheets):
  rows = sheets_read(project.task['auth'], sheets['sheet'], sheets['tab'],
                     'A3:Y')
  return rows_to_videos(rows)


def videos_from_bigquery(bigquery):
  rows = query_to_rows(project.task['auth'],
                       bigquery.get('project', project.id), bigquery['dataset'],
                       'SELECT * FROM %s;' % bigquery['table'])
  return rows_to_videos(rows)


@project.from_parameters
def video():
  if project.verbose:
    print('Video')

  videos = []

  if project.task.get('sheets', {}).get('tab'):
    videos.extend(videos_from_sheets(project.task['sheets']))

  if project.task.get('bigquery', {}).get('table'):
    videos.extend(videos_from_bigquery(project.task['bigquery']))

  if project.task.get('file_or_url'):
    videos.append(project.task)

  for video in videos:
    clip = edit_video(video)
    save_video(video['out'], clip)


if __name__ == '__main__':
  video()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import argparse
import textwrap

from starthinker.util.project import project
from starthinker.util.dbm import lineitem_read


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Command line interface for fetching line items via legacy API.
      Helps developers quickly debug lineitem issues or permissions access issues.

      Example: python helper.py [line item id] -u [user credentials]
  """))

  parser = argparse.ArgumentParser()
  parser.add_argument(
      'lineitem', help='lineitem ID to pull schema, or "list" to get index')

  # initialize project
  project.from_commandline(parser=parser, arguments=('-u', '-c', '-s', '-v'))
  auth = 'service' if project.args.service else 'user'

  # print lineitem
  for row in lineitem_read(auth, lineitems=[project.args.lineitem]):
    print(row)


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.google_api import API_DCM
from starthinker.util.dcm import get_profile_for_api
from starthinker.util.project import project, get_project


def bulkdozer_test():
  print('testing bulkdozer')

  if 'verify' in project.task['traffic']:
    is_admin, profile_id = get_profile_for_api(
        project.task['auth'], project.task['traffic']['account_id'])

    for entity in project.task['traffic']['verify']:
      service = getattr(
          API_DCM(project.task['auth'], internal=is_admin), entity['type'])
      cm_entity = service().get(profileId=profile_id, id=entity['id']).execute()

      values = entity['values']

      for key in values:
        if values[key] != cm_entity[key]:
          raise '%s %s expected to be %s, was %s' % (entity['type'], key, values[key], cm_entity[key])



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of Ads."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.feed import FieldMap


class EventTagDAO(BaseDAO):
  """Event Tag data access object.

  Inherits from BaseDAO and implements Event Tag specific logic for creating and
  updating Event Tags.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes EventTagDAO with profile id and authentication scheme."""
    super(EventTagDAO, self).__init__(auth, profile_id, is_admin)

    self._id_field = FieldMap.EVENT_TAG_ID

    # This causes the dao to search event tag by name, but
    # to do so it is required to pass campaign or advertiser id
    # self._search_field = FieldMap.EVENT_TAG_NAME
    #self._search_field = FieldMap.EVENT_TAG_NAME
    self._search_field = None

    self._list_name = 'eventTags'
    self._entity = 'EVENT_TAGS'
    self._parent_filter_name = 'advertiserId'
    self._parent_dao = None
    self._parent_filter_field_name = FieldMap.ADVERTISER_ID
    self._campaign_dao = CampaignDAO(auth, profile_id, is_admin)

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(EventTagDAO, self)._api(iterate).eventTags()

  def pre_fetch(self, feed):
    """Pre-fetches all required items to be update into the cache.

    This increases performance for update operations.

    Args:
      feed: List of feed items to retrieve
    """
    pass

  def _get_base_search_args(self, search_string):
    return {
        'profileId': self.profile_id,
        'searchString': search_string,
        'sortField': 'NAME'
    }

  def _process_update(self, item, feed_item):
    """Processes the update of an Event Tag

    Args:
      item: Object representing the event tag to be updated, this object is
        updated directly.
      feed_item: Feed item representing event tag values from the Bulkdozer
        feed.
    """
    item['name'] = feed_item.get(FieldMap.EVENT_TAG_NAME, None)
    item['status'] = feed_item.get(FieldMap.EVENT_TAG_STATUS, None)
    item['type'] = feed_item.get(FieldMap.EVENT_TAG_TYPE, None)
    item['url'] = feed_item.get(FieldMap.EVENT_TAG_URL, None)

  def _process_new(self, feed_item):
    """Creates a new event tag DCM object from a feed item representing a event tag from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the event tag from the Bulkdozer feed.

    Returns:
      A event tag object ready to be inserted in DCM through the API.

    """
    campaign = self._campaign_dao.get(feed_item, required=True)

    return {
        'advertiserId':
            feed_item.get(FieldMap.ADVERTISER_ID, None),
        'campaignId':
            campaign.get('id', None) if campaign else None,
        'enabledByDefault':
            feed_item.get(FieldMap.EVENT_TAG_ENABLED_BY_DEFAULT, False),
        'name':
            feed_item.get(FieldMap.EVENT_TAG_NAME, None),
        'status':
            feed_item.get(FieldMap.EVENT_TAG_STATUS, None),
        'type':
            feed_item.get(FieldMap.EVENT_TAG_TYPE, None),
        'url':
            feed_item.get(FieldMap.EVENT_TAG_URL, None)
    }

  def _post_process(self, feed_item, item):
    """Updates the feed item with ids and names of related object so those can be updated in the Bulkdozer feed.

    Args:
      feed_item: The Bulkdozer feed item.
      item: The CM newly created or updated object.
    """
    campaign = self._campaign_dao.get(feed_item, required=True)

    if campaign:
      feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']
      feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
import argparse
import textwrap

from starthinker.util.project import project
from starthinker.util.auth import get_profile


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Creates USER credentials from Google Cloud Project CLIENT Credentials and displays profile information if it worked.
      CLIENT credentials are required to run this script, to obtain the JSON file...

        Step 1: Configure Authentication Consent Screen ( do only once )
        ----------------------------------------
          A. Visit: https://console.developers.google.com/apis/credentials/consent
          B. Choose Internal if you have GSuite, otherwise choose External.
          C. For Application Name enter: StarThinker
          D. All other fields are optional, click Save.

        Step 2: Create CLIENT Credentials ( do only once )
        ----------------------------------------
          A. Visit: https://console.developers.google.com/apis/credentials/oauthclient
          B. Choose Desktop.
          C. For Name enter: StarThinker.
          D. Click Create and ignore the confirmation pop-up.

        Step 3: Download CLIENT Credentials File ( do only once )"
        ----------------------------------------"
          A. Visit: https://console.developers.google.com/apis/credentials"
          B. Find your newly created key under OAuth 2.0 Client IDs and click download arrow on the right."
          C. The downloaded file is the CLIENT credentials, use its path for the --client -c parameter.

        Step 4: Generate USER Credentials File ( do only once )"
        ----------------------------------------"
          A. Run this command with parameters -c [CLIENT file path] and -u [USER file path].
          B. The USER file will be created and can be used to access Google APIs.
          C. The user profile will be printed to the screen

        Example: python helper.py -u [CLIENT file path] -c [USER file path]
"""))

  # initialize project, enable only some default arguments
  project.from_commandline(parser=parser, arguments=('-c', '-u'))

  # get profile to verify everything worked
  print('Profile:', json.dumps(get_profile(), indent=2, sort_keys=True))


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

DV-3PO Custom Signals

DV-3PO Custom Signals allows automated changes to be made to DV360 campaigns based on external signals from weather and social media trends. In the future it will also support news, disaster alerts, stocks, sports, custom APIs, etc.

  - Open the template sheet: <a target='_blank' href='https://docs.google.com/spreadsheets/d/1xl2bknsn9ptCKwg1sZqTBpmRFDdcG3ATdp2f4C2CQ2U/edit#gid=1579485492'>[DV-3PO] Custom Signals Configs</a>.
  - Make a copy of the sheet through the menu File -> Make a copy, for clarity we suggest you rename the copy to a meaningful name describing the usage of this copy.
  - In the Station IDs field below enter a comma separated list of NOAA weather station IDs. Most major airports are stations and their ID typically is K followed by the 3 letter airport code, e.g. KORD for Chicago O'Hare International Airport, KSFO for San Francisco international airport, etc. You can get a full list of stations <a target='_blank' href='https://www1.ncdc.noaa.gov/pub/data/noaa/isd-history.txt'>here</a>, the station ID to use is the 'CALL' column of this list.
  - In the Sheet URL field below, enter the URL of the copy of the config sheet you've created.
  - Go to the sheet and configure the rules you'd like to be applied in the Rules tab.
  - In the Advertiser ID column, enter the advertiser ID of the line items you'd like to automatically update.
  - In the Line Item ID colunn, enter the line item ID of the line item you'd like to automatically update.
  - The 'Active' column of the Rules tab allows you to control if the line item should be active or paused. If this field is TRUE the line item will be set to active, if this field is FALSE the line item will be set to inactive. You can use a formula to take weather data into consideration to update this field, e.g. =IF(Weather!C2>30, TRUE, FALSE) will cause the line item to be activated if the temperature of the first station in the Weather tab is above 30 degrees. Leave this field empty if you don't want it to be modified by the tool.
  - The 'Fixed Bid' column of the Rules tab allows you to control the fixed bid amount of the line item. The value set to this field will be applied to the specified line item. You can use a formula to take weather data into consideration to update this field, e.g. =IF(Weather!G2>3, 0.7, 0.4) will cause bid to be set to $0.7 if the wind speed of the first line in the Weather tab is greater than 3 mph, or $0.4 otherwise. Leave this field empty if you don't want it to be modified by the tool.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'station_ids': '',  # NOAA Weather Station ID
  'auth_read': 'user',  # Credentials used for reading data.
  'sheet_url': '',  # Feed Sheet URL
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      0,
      1,
      2,
      3,
      4,
      5,
      6,
      7,
      8,
      9,
      10,
      11,
      12,
      13,
      14,
      15,
      16,
      17,
      18,
      19,
      20,
      21,
      22,
      23
    ]
  },
  'tasks': [
    {
      'weather_gov': {
        'auth': 'user',
        'stations': {
          'field': {
            'name': 'station_ids',
            'kind': 'string_list',
            'order': 1,
            'description': 'NOAA Weather Station ID',
            'default': ''
          }
        },
        'out': {
          'sheets': {
            'sheet': {
              'field': {
                'name': 'sheet_url',
                'kind': 'string',
                'order': 2,
                'description': 'Feed Sheet URL',
                'default': ''
              }
            },
            'tab': 'Weather',
            'range': 'A2:K',
            'delete': True
          }
        }
      }
    },
    {
      'lineitem_beta': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'read': {
          'sheet': {
            'sheet': {
              'field': {
                'name': 'sheet_url',
                'kind': 'string',
                'order': 2,
                'description': 'Feed Sheet URL',
                'default': ''
              }
            },
            'tab': 'Rules',
            'range': 'A1:D'
          }
        },
        'patch': {
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('dv3po_custom_signals', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Transparency Dashboard

Reports the percentage of CM impressions that can be attributed to a specific domain or application.  Allows diagnostic of which domains and apps are misconfigured by publisher resulting in underreporting.

  - Wait for <a href='https://console.cloud.google.com/bigquery?project=UNDEFINED&d=UNDEFINED' target='_blank'>BigQuery : UNDEFINED</a> : UNDEFINED : UNDEFINED</a> to be created.
  - Copy DataStudio <a href='https://datastudio.google.com/c/u/0/datasources/1Az6d9loAHo69GSIyKUfusrtyf_IDqTVs' target='_blank'>Transparency Combined KPI</a> and connect.
  - Copy DataStudio <a href='https://datastudio.google.com/c/u/0/reporting/1foircGRxgYCL_PR8gfdmYOleBacnPKwB/page/QCXj' target='_blank'>Transparency Breakdown</a>.
  - When prompted choose the new data source you just created.
  - Or give these intructions to the client, they will have to join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a>.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be written in BigQuery.
  'recipe_name': '',  # Name of report in CM, should be unique.
  'recipe_project': '',  # Project where BigQuery dataset will be created.
  'dcm_account': '',  # CM account id of client.
  'dcm_advertisers': '',  # Comma delimited list of CM advertiser ids.
}

RECIPE = {
  'tasks': [
    {
      'dataset': {
        'hour': [
          1
        ],
        'auth': 'service',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 1,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        }
      }
    },
    {
      'dcm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'account': {
            'field': {
              'name': 'dcm_account',
              'kind': 'integer',
              'order': 2,
              'default': '',
              'description': 'CM account id of client.'
            }
          },
          'filters': {
            'dfa:advertiser': {
              'values': {
                'field': {
                  'name': 'dcm_advertisers',
                  'kind': 'integer_list',
                  'order': 3,
                  'description': 'Comma delimited list of CM advertiser ids.'
                }
              }
            }
          },
          'body': {
            'type': 'STANDARD',
            'format': 'CSV',
            'name': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'Transparency App For ',
                'description': 'Name of report in CM, unique.'
              }
            },
            'criteria': {
              'dateRange': {
                'relativeDateRange': 'PREVIOUS_MONTH'
              },
              'dimensions': [
                {
                  'name': 'dfa:advertiser'
                },
                {
                  'name': 'dfa:advertiserId'
                },
                {
                  'name': 'dfa:campaign'
                },
                {
                  'name': 'dfa:campaignId'
                },
                {
                  'name': 'dfa:siteId'
                },
                {
                  'name': 'dfa:site'
                },
                {
                  'name': 'dfa:adType'
                },
                {
                  'name': 'dfa:environment'
                },
                {
                  'name': 'dfa:appId'
                },
                {
                  'name': 'dfa:app'
                }
              ],
              'metricNames': [
                'dfa:impressions'
              ]
            },
            'schedule': {
              'active': True,
              'every': 1,
              'repeats': 'MONTHLY',
              'runsOnDayOfMonth': 'DAY_OF_MONTH'
            }
          }
        }
      }
    },
    {
      'dcm': {
        'hour': [
          2
        ],
        'auth': 'user',
        'report': {
          'account': {
            'field': {
              'name': 'dcm_account',
              'kind': 'integer',
              'order': 2,
              'default': '',
              'description': 'CM account id of client.'
            }
          },
          'filters': {
            'dfa:advertiser': {
              'values': {
                'field': {
                  'name': 'dcm_advertisers',
                  'kind': 'integer_list',
                  'order': 3,
                  'description': 'Comma delimited list of CM advertiser ids.'
                }
              }
            }
          },
          'body': {
            'type': 'STANDARD',
            'format': 'CSV',
            'name': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'Transparency Domain For ',
                'description': 'Name of report in CM, unique.'
              }
            },
            'criteria': {
              'dateRange': {
                'relativeDateRange': 'PREVIOUS_MONTH'
              },
              'dimensions': [
                {
                  'name': 'dfa:advertiser'
                },
                {
                  'name': 'dfa:advertiserId'
                },
                {
                  'name': 'dfa:campaign'
                },
                {
                  'name': 'dfa:campaignId'
                },
                {
                  'name': 'dfa:site'
                },
                {
                  'name': 'dfa:siteId'
                },
                {
                  'name': 'dfa:adType'
                },
                {
                  'name': 'dfa:domain'
                }
              ],
              'metricNames': [
                'dfa:verificationVerifiableImpressions'
              ]
            },
            'schedule': {
              'active': True,
              'every': 1,
              'repeats': 'MONTHLY',
              'runsOnDayOfMonth': 'DAY_OF_MONTH'
            }
          }
        }
      }
    },
    {
      'dcm': {
        'hour': [
          4
        ],
        'auth': 'user',
        'report': {
          'account': {
            'field': {
              'name': 'dcm_account',
              'kind': 'integer',
              'order': 2,
              'default': '',
              'description': 'CM account id of client.'
            }
          },
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Transparency Domain For ',
              'description': 'Name of report in CM, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'auth': 'service',
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'Transparency_Domain_KPI'
          }
        }
      }
    },
    {
      'dcm': {
        'hour': [
          4
        ],
        'auth': 'user',
        'report': {
          'account': {
            'field': {
              'name': 'dcm_account',
              'kind': 'integer',
              'order': 2,
              'default': '',
              'description': 'CM account id of client.'
            }
          },
          'name': {
            'field': {
              'name': 'recipe_name',
              'kind': 'string',
              'prefix': 'Transparency App For ',
              'description': 'Name of report in CM, should be unique.'
            }
          }
        },
        'out': {
          'bigquery': {
            'auth': 'service',
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 1,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'Transparency_App_KPI'
          }
        }
      }
    },
    {
      'bigquery': {
        'hour': [
          5
        ],
        'auth': 'user',
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 1,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'view': 'Transparency_Combined_KPI'
        },
        'from': {
          'query': "With \r\nTransparent_Domains AS ( \r\n  SELECT\r\n    CONCAT(Advertiser, ' - ', CAST(Advertiser_Id AS STRING)) AS Advertiser,\r\n    CONCAT(Campaign, ' - ', CAST(Campaign_Id AS STRING)) AS Campaign,\r\n    CONCAT(Site_Dcm, ' - ', CAST(Site_Id_Dcm AS STRING)) AS Site,\r\n    Domain,\r\n    Ad_Type,\r\n    Verifiable_Impressions AS Impressions,\r\n    IF(Domain IS NOT NULL, Verifiable_Impressions, 0) AS Visible_Impressions,\r\n    IF(Domain IS NULL, Verifiable_Impressions, 0) AS Null_Impressions\r\n  FROM `[PARAMETER].[PARAMETER].Transparency_Domain_KPI`\r\n),\r\nTransparent_Apps AS ( \r\n  SELECT\r\n    CONCAT(Advertiser, ' - ', CAST(Advertiser_Id AS STRING)) AS Advertiser,\r\n    CONCAT(Campaign, ' - ', CAST(Campaign_Id AS STRING)) AS Campaign,\r\n    CONCAT(Site_Dcm, ' - ', CAST(Site_Id_Dcm AS STRING)) AS Site,\r\n    /*If(App IS NOT NULL, CONCAT(App, ' - ', CAST(App_Id AS STRING)), App_Id) AS App, */\r\n    CASE \r\n      WHEN App IS NOT NULL THEN CONCAT(App, ' - ', CAST(App_Id AS STRING))\r\n      WHEN App_Id IS NOT NULL THEN App_Id\r\n      ELSE NULL\r\n    END AS App,\r\n    Ad_Type,\r\n    Impressions,\r\n    IF(App IS NOT NULL OR App_ID IS NOT NULL, Impressions, 0) AS Visible_Impressions,\r\n    IF(App IS NULL AND App_Id IS NULL, Impressions, 0) AS Null_Impressions\r\n  FROM `[PARAMETER].[PARAMETER].Transparency_App_KPI`\r\n  WHERE Environment = 'App'\r\n),\r\nDomains_And_Apps AS (\r\n  SELECT \r\n    TD.Advertiser,\r\n    TD.Campaign,\r\n    TD.Site,\r\n    TD.Ad_Type,\r\n    TD.Domain,\r\n    TD.Impressions AS Domain_Impressions,\r\n    TD.Visible_Impressions AS Domain_Visible_Impressions,\r\n    TD.Null_Impressions AS Domain_Null_Impressions,\r\n    NULL AS App,\r\n    0 AS App_Impressions,\r\n    0 AS App_Visible_Impressions,\r\n    0 AS App_Null_Impressions\r\n  FROM Transparent_Domains AS TD\r\n  UNION ALL\r\n  SELECT \r\n    TA.Advertiser,\r\n    TA.Campaign,\r\n    TA.Site,\r\n    TA.Ad_Type,\r\n    NULL AS Domain,\r\n    0 AS Domain_Impressions,\r\n    0 AS Domain_Visible_Impressions,\r\n    0 AS Domain_Null_Impressions,\r\n    TA.App,\r\n    TA.Impressions AS App_Impressions,\r\n    TA.Visible_Impressions AS App_Visible_Impressions,\r\n    TA.Null_Impressions AS App_Null_Impressions\r\n  FROM Transparent_Apps AS TA\r\n)\r\n\r\n  SELECT\r\n    Advertiser,\r\n    Campaign,\r\n    Site,\r\n    COALESCE(Domain, App, '') AS Domain_Or_App,\r\n    Ad_Type,\r\n    CASE\r\n      WHEN App IS NOT NULL AND Domain IS NOT NULL THEN 'Both' /* SHOULD NOT HAPPEN */\r\n      WHEN App IS NOT NULL THEN 'App'\r\n      WHEN Domain IS NOT NULL Then 'Domain'\r\n      ELSE 'Neither'\r\n    END AS Category,\r\n\r\n    SUM(Domain_Impressions) AS Domain_Impressions,\r\n    SUM(Domain_Visible_Impressions) AS Domain_Visible_Impressions,\r\n    SUM(Domain_Null_Impressions) AS Domain_Null_Impressions,\r\n\r\n    SUM(App_Impressions) AS App_Impressions,\r\n    SUM(App_Visible_Impressions) AS App_Visible_Impressions,\r\n    SUM(App_Null_Impressions) AS App_Null_Impressions,\r\n\r\n    SUM(App_Impressions + Domain_Impressions) AS Impressions /* Could also be MAX as its always one or the other*/\r\n\r\n  FROM Domains_And_Apps\r\n  GROUP By 1,2,3,4,5,6",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ],
          'legacy': False
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('transparency', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Federal Reserve Regional Data

Download federal reserve region.

  - Specify the values for a <a href='https://research.stlouisfed.org/docs/api/geofred/regional_data.html' target='_blank'>Fred observations API call</a>.
  - A table will appear in the dataset.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth': 'service',  # Credentials used for writing data.
  'fred_api_key': '',  # 32 character alpha-numeric lowercase string.
  'fred_series_group': '',  # The ID for a group of seriess found in GeoFRED.
  'fred_region_type': 'county',  # The region you want want to pull data for.
  'fred_units': 'lin',  # A key that indicates a data value transformation.
  'fred_frequency': '',  # An optional parameter that indicates a lower frequency to aggregate values to.
  'fred_season': 'SA',  # The seasonality of the series group.
  'fred_aggregation_method': 'avg',  # A key that indicates the aggregation method used for frequency aggregation.
  'project': '',  # Existing BigQuery project.
  'dataset': '',  # Existing BigQuery dataset.
}

RECIPE = {
  'tasks': [
    {
      'fred': {
        'auth': {
          'field': {
            'name': 'auth',
            'kind': 'authentication',
            'order': 0,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'api_key': {
          'field': {
            'name': 'fred_api_key',
            'kind': 'string',
            'order': 1,
            'default': '',
            'description': '32 character alpha-numeric lowercase string.'
          }
        },
        'frequency': {
          'field': {
            'name': 'fred_frequency',
            'kind': 'choice',
            'order': 4,
            'default': '',
            'description': 'An optional parameter that indicates a lower frequency to aggregate values to.',
            'choices': [
              '',
              'd',
              'w',
              'bw',
              'm',
              'q',
              'sa',
              'a',
              'wef',
              'weth',
              'wew',
              'wetu',
              'wem',
              'wesu',
              'wesa',
              'bwew',
              'bwem'
            ]
          }
        },
        'region_type': {
          'field': {
            'name': 'fred_region_type',
            'kind': 'choice',
            'order': 3,
            'default': 'county',
            'description': 'The region you want want to pull data for.',
            'choices': [
              'bea',
              'msa',
              'frb',
              'necta',
              'state',
              'country',
              'county',
              'censusregion'
            ]
          }
        },
        'regions': [
          {
            'series_group': {
              'field': {
                'name': 'fred_series_group',
                'kind': 'string',
                'order': 2,
                'default': '',
                'description': 'The ID for a group of seriess found in GeoFRED.'
              }
            },
            'units': {
              'field': {
                'name': 'fred_units',
                'kind': 'choice',
                'order': 3,
                'default': 'lin',
                'description': 'A key that indicates a data value transformation.',
                'choices': [
                  'lin',
                  'chg',
                  'ch1',
                  'pch',
                  'pc1',
                  'pca',
                  'cch',
                  'cca',
                  'log'
                ]
              }
            },
            'season': {
              'field': {
                'name': 'fred_season',
                'kind': 'choice',
                'order': 4,
                'default': 'SA',
                'description': 'The seasonality of the series group.',
                'choices': [
                  'SA',
                  'NSA',
                  'SSA'
                ]
              }
            },
            'aggregation_method': {
              'field': {
                'name': 'fred_aggregation_method',
                'kind': 'choice',
                'order': 5,
                'default': 'avg',
                'description': 'A key that indicates the aggregation method used for frequency aggregation.',
                'choices': [
                  'avg',
                  'sum',
                  'eop'
                ]
              }
            }
          }
        ],
        'out': {
          'bigquery': {
            'project': {
              'field': {
                'name': 'project',
                'kind': 'string',
                'order': 10,
                'default': '',
                'description': 'Existing BigQuery project.'
              }
            },
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 11,
                'default': '',
                'description': 'Existing BigQuery dataset.'
              }
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('fred_regional_to_bigquery', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer from open source:

    pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory:

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

DV360 API To BigQuery

Write the current state of advertisers, campaigns, sites, insertion orders, and line items to BigQuery for a given list of DV360 accounts.

Specify the name of the dataset, several tables will be created here.
If dataset exists, it is unchanged.
Add DV360 advertiser ids for the accounts to pull data from.

'''

from starthinker_airflow.factory import DAG_Factory

# Add the following credentials to your Airflow configuration.
USER_CONN_ID = "starthinker_user" # The connection to use for user authentication.
GCP_CONN_ID = "starthinker_service" # The connection to use for service authentication.

INPUTS = {
  'endpoint': '',
  'auth_read': 'user',  # Credentials used for reading data.
  'auth_write': 'service',  # Credentials used for writing data.
  'dataset': '',  # Google BigQuery dataset to create tables in.
  'partners': [],  # Comma separated partners ids.
  'advertisers': [],  # Comma separated advertisers ids.
}

TASKS = [
  {
    'dv360_api': {
      'auth': {
        'field': {
          'name': 'auth_read',
          'kind': 'authentication',
          'order': 1,
          'default': 'user',
          'description': 'Credentials used for reading data.'
        }
      },
      'endpoints': {
        'field': {
          'name': 'endpoint',
          'kind': 'choice',
          'choices': [
            'advertisers',
            'advertisers.campaigns',
            'advertisers.channels',
            'advertisers.channels.sites',
            'advertisers.creatives',
            'advertisers.insertionOrders',
            'advertisers.lineItems',
            'advertisers.lineItems.targetingTypes',
            'advertisers.locationLists',
            'advertisers.locationLists.assignedLocations',
            'advertisers.negativeKeywordLists',
            'advertisers.negativeKeywordLists.negativeKeywords',
            'advertisers.targetingTypes.assignedTargetingOptions',
            'combinedAudiences',
            'customBiddingAlgorithms',
            'customLists',
            'firstAndThirdPartyAudiences',
            'floodlightGroups',
            'googleAudiences',
            'inventorySourceGroups',
            'inventorySourceGroups.assignedInventorySources',
            'inventorySources',
            'partners',
            'partners.channels',
            'partners.channels.sites',
            'floodlightActivityGroups',
            'partners.targetingTypes.assignedTargetingOptions',
            'targetingTypes.targetingOptions',
            'users'
          ],
          'default': ''
        }
      },
      'partners': {
        'single_cell': True,
        'values': {
          'field': {
            'name': 'partners',
            'kind': 'integer_list',
            'order': 2,
            'default': [
            ],
            'description': 'Comma separated partners ids.'
          }
        }
      },
      'advertisers': {
        'single_cell': True,
        'values': {
          'field': {
            'name': 'advertisers',
            'kind': 'integer_list',
            'order': 2,
            'default': [
            ],
            'description': 'Comma separated advertisers ids.'
          }
        }
      },
      'out': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 1,
            'default': '',
            'description': 'Google BigQuery dataset to create tables in.'
          }
        }
      }
    }
  }
]

DAG_FACTORY = DAG_Factory('dv360_api_to_bigquery', { 'tasks':TASKS }, INPUTS)
DAG_FACTORY.apply_credentails(USER_CONN_ID, GCP_CONN_ID)
DAG = DAG_FACTORY.execute()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.csv import rows_pad
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.line_item import line_item_commit
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def line_item_map_clear():
  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Line Items Map',
               'A2:Z')


def line_item_map_load():
  pass


def line_item_map_audit():
  rows = get_rows(
      project.task['auth_sheets'], {
          'sheets': {
              'sheet': project.task['sheet'],
              'tab': 'Line Items Map',
              'range': 'A2:Z'
          }
      })

  put_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset': project.task['dataset'],
              'table': 'SHEET_LineItemMaps',
              'schema': [{
                  'name': 'Action',
                  'type': 'STRING'
              }, {
                  'name': 'Line_Item',
                  'type': 'STRING'
              }, {
                  'name': 'Creative',
                  'type': 'STRING'
              }],
              'format': 'CSV'
          }
      }, rows)

  query_to_view(
      project.task['auth_bigquery'],
      project.id,
      project.task['dataset'],
      'AUDIT_LineItemMaps',
      """WITH
      LINEITEM_ERRORS AS (
      SELECT
        'Line Items Map' AS Operation,
        'Missing Line Item.' AS Error,
        'ERROR' AS Severity,
        COALESCE(Line_Item, 'BLANK') AS Id
      FROM
        `{dataset}.SHEET_LineItemMaps` AS M
      LEFT JOIN
        `{dataset}.DV_LineItems` AS L
      ON
        M.Line_Item=CONCAT(L.displayName, ' - ', L.lineItemId)
      WHERE L IS NULL
      ),
      CREATIVE_ERRORS AS (
      SELECT
        'Line Items Map' AS Operation,
        'Missing Line Item.' AS Error,
        'ERROR' AS Severity,
        COALESCE(Line_Item, 'BLANK') AS Id
      FROM
        `{dataset}.SHEET_LineItemMaps` AS M
      LEFT JOIN
        `{dataset}.DV_Creatives` AS C
      ON
        M.Line_Item=CONCAT(C.displayName, ' - ', C.creativeId)
      WHERE C IS NULL
      )
      SELECT * FROM LINEITEM_ERRORS
      UNION ALL
      SELECT * FROM CREATIVE_ERRORS
      ;
    """.format(**project.task),
      legacy=False)


def line_item_map_patch(commit=False):
  patches = {}
  changed = set()

  rows = get_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'query':
                  """SELECT advertiserId, lineItemId, creativeIds FROM `{dataset}.DV_LineItems`
      """.format(**project.task),
              'as_object':
                  True,
              'legacy':
                  False
          }
      })

  for row in rows:
    patches[str(row['lineItemId'])] = {
        'operation': 'Line Items Map',
        'action': 'PATCH',
        'parameters': {
            'advertiserId': str(row['advertiserId']),
            'lineItemId': str(row['lineItemId']),
            'body': {
                'creativeIds': [str(c) for c in row['creativeIds']]
            }
        }
    }

  rows = get_rows(
      project.task['auth_sheets'], {
          'sheets': {
              'sheet': project.task['sheet'],
              'tab': 'Line Items Map',
              'range': 'A2:Z'
          }
      })

  rows = rows_pad(rows, 3, '')

  for row in rows:
    lineitem_id = lookup_id(row[1])
    creative_id = lookup_id(row[2])
    if lineitem_id in patches:
      if row[0] == 'ADD' and creative_id not in patches[lineitem_id][
          'parameters']['body']['creativeIds']:
        patches[lineitem_id]['line_item'] = row[1]
        patches[lineitem_id]['parameters']['body']['creativeIds'].append(
            creative_id)
        changed.add(lineitem_id)
      if row[0] == 'REMOVE' and creative_id in patches[lineitem_id][
          'parameters']['body']['creativeIds']:
        patches[lineitem_id]['line_item'] = row[1]
        patches[lineitem_id]['parameters']['body']['creativeIds'].remove(
            creative_id)
        changed.add(lineitem_id)

  # Remove any patches where creatives have not changed
  for li in list(patches.keys()):
    if li not in changed:
      del patches[li]
  patches = list(patches.values())

  patch_masks(patches)

  if commit:
    line_item_commit(patches)
  else:
    patch_preview(patches)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.project import project
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.bid_strategy import bid_strategy_audit
from starthinker.task.dv_editor.frequency_cap import frequency_cap_audit
from starthinker.task.dv_editor.integration_detail import integration_detail_audit
from starthinker.task.dv_editor.insertion_order import insertion_order_audit
from starthinker.task.dv_editor.line_item import line_item_audit
from starthinker.task.dv_editor.line_item_map import line_item_map_audit
from starthinker.task.dv_editor.pacing import pacing_audit
from starthinker.task.dv_editor.partner_cost import partner_cost_audit
from starthinker.task.dv_editor.segment import segment_audit


def audit_clear():
  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Audit', 'A2:Z')


def audit_load():

  bid_strategy_audit()
  integration_detail_audit()
  insertion_order_audit()
  frequency_cap_audit()
  line_item_audit()
  line_item_map_audit()
  pacing_audit()
  partner_cost_audit()
  segment_audit()

  # write audit to sheet
  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Audit', 'A2')

  # write audits to sheet
  rows = get_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'query':
                  """SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_InsertionOrders`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_Segments`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_LineItems`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_LineItemMaps`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_Pacing`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_BidStrategy`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_FrequencyCaps`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_PartnerCosts`
      UNION ALL
        SELECT Operation, Severity, Id, Error
        FROM `{dataset}.AUDIT_IntegrationDetails`
      """.format(**project.task),
              'legacy':
                  False
          }
      })

  put_rows(project.task['auth_sheets'], {
      'sheets': {
          'sheet': project.task['sheet'],
          'tab': 'Audit',
          'range': 'A2'
      }
  }, rows)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "dbm":{...}} task in recipe JSON.

This script translates JSON instructions into operations on DBM reporting.
It deletes, or creates, and/or downloads DBM reports.  See JSON files in
this directory for examples of operations.

This script uses put_rows as defined in util/data/README.md. This allows
multiple destinations for downloaded reports. To add a destination modify
the util/data/__init__.py functions.

Note

The underlying libraries use streaming download buffers, no disk is used.
Buffers are controlled in config.py.

"""

from starthinker.util.project import project
from starthinker.util.data import put_rows, get_rows
from starthinker.util.dbm import report_delete, report_filter, report_build, report_file, report_to_rows, report_clean


@project.from_parameters
def dbm():
  if project.verbose:
    print('DBM')

  # name is redundant if title is given, allow skipping use of name for creating reports
  if 'body' in project.task['report'] and 'name' not in project.task['report']:
    project.task['report']['name'] = project.task['report']['body']['metadata'][
        'title']

  # check if report is to be deleted
  if project.task.get('delete', False):
    if project.verbose:
      print('DBM DELETE')
    report_delete(project.task['auth'],
                  project.task['report'].get('report_id', None),
                  project.task['report'].get('name', None))

  # check if report is to be created
  if 'body' in project.task['report']:
    if project.verbose:
      print('DBM BUILD', project.task['report']['body']['metadata']['title'])

    # ceck if filters given ( returns new body )
    if 'filters' in project.task['report']:
      project.task['report']['body'] = report_filter(
          project.task['auth'], project.task['report']['body'],
          project.task['report']['filters'])

    # create the report
    report = report_build(project.task['auth'], project.task['report']['body'])

  # moving a report
  if 'out' in project.task:

    filename, report = report_file(
        project.task['auth'], project.task['report'].get('report_id', None),
        project.task['report'].get('name', None),
        project.task['report'].get('timeout', 10))

    # if a report exists
    if report:
      if project.verbose:
        print('DBM FILE', filename)

      # clean up the report
      rows = report_to_rows(report)
      rows = report_clean(rows)

      # write rows using standard out block in json ( allows customization across all scripts )
      if rows:
        put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  dbm()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import json
from difflib import Differ
from time import sleep

from starthinker.util.project import project
from starthinker.util.bigquery import table_to_rows
from starthinker.util.bigquery import table_to_schema
from starthinker.util.bigquery import query_parameters
from starthinker.util.bigquery import query_to_rows
from starthinker.util.sheets import sheets_read
from starthinker.util.storage import object_exists, object_delete
from starthinker.util.drive import file_exists, file_delete
from starthinker.util.csv import rows_to_type
from starthinker.task.traffic.test import bulkdozer_test
from starthinker.task.weather_gov.test import weather_gov_test


EXIT_ERROR = 1
EXIT_SUCCESS = 0


def test_failed():
  print('FAILED')
  return EXIT_ERROR


def test_passed():
  print('PASSED')
  return EXIT_SUCCESS


def schema_compare(expected, actual, path=''):
  delta = {}
  matched = set()

  make_path = lambda name: path + ('.' if path else '') + name

  # find matches
  for expected_column in expected:
    for actual_column in actual:
      if expected_column['name'] == actual_column['name']:
        matched.add(expected_column['name'])
        if expected_column.get('type') != actual_column.get('type'):
          delta[make_path(expected_column['name'])] = { 'path': make_path(expected_column['name']), 'error': 'type', 'expected': expected_column.get('type'), 'actual': actual_column.get('type')}
        if expected_column.get('mode') != actual_column.get('mode'):
          delta[make_path(expected_column['name'])] = { 'path': make_path(expected_column['name']), 'error': 'mode', 'expected': expected_column.get('mode'), 'actual': actual_column.get('mode')}
        delta.update(schema_compare(expected_column.get('fields', []), actual_column.get('fields', []), make_path(expected_column['name'])))

  # find missing
  for expected_column in expected:
    if expected_column['name'] not in matched:
      delta[make_path(expected_column['name'])] = { 'path': make_path(expected_column['name']), 'error': 'missing', 'expected': expected_column['name'], 'actual': ''}

  # find extra
  for actual_column in actual:
    if actual_column['name'] not in matched:
      delta[make_path(actual_column['name'])] = { 'path': make_path(actual_column['name']), 'error': 'extra', 'expected': '', 'actual': actual_column['name']}

  return delta


def deep_compare(actual, expected):

  if type(actual) != type(expected):
    return 'EXPECTED %s BUT ACTUAL %s' % (type(expected), type(actual))

  elif isinstance(expected, (dict, tuple, list)):
    expected_str = json.dumps(expected, indent=2, sort_keys=True, default=str)
    actual_str = json.dumps(actual, indent=2, sort_keys=True, default=str)

    delta = list(Differ().compare(expected_str.splitlines(),
                                  actual_str.splitlines()))

    if sum(1 for d in delta if d[0] in ['-', '+', '?']):
      return ('%s\nEXPECTED '
              '*******************************************************\n%s\nACTUAL'
              ' *******************************************************\n%s') % (
          '\n'.join(delta), expected_str, actual_str)

  elif actual != expected:
    return 'EXPECTED %s != ACTUAL %s' % (expected, actual)

  return None


# display results of list comparison
def object_compare(actual, expected):
  errors = deep_compare(actual, expected)

  if errors:
    print('\nOBJECT DELTA *************************************************\n')
    print(errors)
    print('\n**************************************************************\n')
    test_failed()
  else:
    test_passed()


# check if sheet matches given values
#    { "test": {
#      "auth":"user",
#      "sheets": {
#        "sheet":"https://docs.google.com/spreadsheets/d/1h-Ic-DlCv-Ct8-k-VJnpo_BAkqsS70rNe0KKeXKJNx0/edit?usp=sharing",
#        "tab":"Sheet_Clear",
#        "range":"A1:C",
#        "values":[
#          ["Animal", "Age", "Weight ( lbs )"],
#          ["dog", 7, 67],
#          ["cat", 5, 1.5],
#          ["bird", 12, 0.44],
#          ["lizard", 22, 1],
#          ["dinosaur", 1600, 273.97]
#        ]
#      }
#    }}


def sheets():
  print('TEST: sheets')

  rows = sheets_read(project.task['auth'], project.task['sheets']['sheet'],
                     project.task['sheets']['tab'],
                     project.task['sheets']['range'])

  rows = rows_to_type(rows)
  object_compare(list(rows), project.task['sheets']['values'])


# check if bigquery table has given values or has data
#    { "test": {
#      "auth":"user",
#      "bigquery":{
#        "dataset":"Test",
#        "table":"Sheet_To_BigQuery",
#        "schema":[
#          {"name": "Animal", "type": "STRING"},
#          {"name": "Age", "type": "INTEGER"},
#          {"name": "Weight_lbs", "type": "FLOAT"}
#        ],
#        "values":[
#          ["dog", 7, 67],
#          ["cat", 5, 1.5],
#          ["bird", 12, 0.44],
#          ["lizard", 22, 1],
#          ["dinosaur", 1600, 273.97]
#        ]
#      }
#    }}


def bigquery():
  print('TEST: bigquery')

  # check schema if given ( check independent of values )
  if 'schema' in project.task['bigquery']:
    schema = table_to_schema(project.task['auth'], project.id,
                             project.task['bigquery']['dataset'],
                             project.task['bigquery']['table'])
    deltas = schema_compare(project.task['bigquery']['schema'], schema, path='')

    if deltas:
      print('\nFAILED *******************************************************\n')
      for delta in deltas.values():
        print('%(path)s: %(error)s ( %(expected)s - %(actual)s)' % delta)
      print('\n**************************************************************\n')
      test_failed()
    else:
      test_passed()


  # if query given check it
  if 'query' in project.task['bigquery']:
    rows = query_to_rows(
        project.task['auth'],
        project.id,
        project.task['bigquery']['dataset'],
        query_parameters(project.task['bigquery']['query'],
                         project.task['bigquery'].get('parameters')),
    )

    object_compare(sorted(rows), sorted(project.task['bigquery']['values']))

  # simple table check ( unless query given )
  elif 'values' in project.task['bigquery']:
    rows = table_to_rows(project.task['auth'], project.id,
                         project.task['bigquery']['dataset'],
                         project.task['bigquery']['table'])

    object_compare(sorted(rows), sorted(project.task['bigquery']['values']))


def asserts():
  print('TEST: asserts')
  print(project.task['assert'])
  test_passed()


def path_exists():
  print('TEST: path_exists')
  if os.path.exists(project.task['path']):
    if project.task.get('delete', False):
      os.remove(project.task['path'])
    test_passed()
  else:
    test_failed()


def storage_exists():
  print('TEST: storage_exists')
  if object_exists(
      project.task['auth'], '%s:%s' %
      (project.task['storage']['bucket'], project.task['storage']['file'])):
    if project.task.get('delete', False):
      object_delete(
          project.task['auth'], '%s:%s' %
          (project.task['storage']['bucket'], project.task['storage']['file']))
    test_passed()
  else:
    test_failed()


def drive_exists():
  print('TEST: drive')
  if file_exists(project.task['auth'], project.task['drive']['file']):
    if project.task.get('delete', False):
      file_delete(project.task['auth'], project.task['drive']['file'])
    test_passed()
  else:
    test_failed()


def test_sleep():
  print('TEST: sleep ', project.task['sleep'])
  sleep(project.task['sleep'])


def weather_gov():
  print('TEST: weather_gov')
  try:
    weather_gov_test()
    test_passed()
  except Exception as e:
    print(str(e))
    test_failed()


def traffic():
  print('TEST: Bulkdozer')
  try:
    bulkdozer_test()
    test_passed()
  except Exception as e:
    print(str(e))
    test_failed()


# decide which test to run
@project.from_parameters
def test():
  if 'assert' in project.task:
    return asserts()
  elif 'path' in project.task:
    return path_exists()
  elif 'storage' in project.task:
    storage_exists()
  elif 'sheets' in project.task:
    return sheets()
  elif 'bigquery' in project.task:
    return bigquery()
  elif 'drive' in project.task:
    return drive_exists()
  elif 'sleep' in project.task:
    return test_sleep()
  elif 'template' in project.task:
    return template()
  elif 'traffic' in project.task:
    return traffic()
  elif 'weather_gov' in project.task:
    return weather_gov()


# test should be run like any other task
# one test per task ( otherwise it gets confusing )
# calling script already indicates which test is being run
# print only PASS or FAIL
if __name__ == '__main__':
  test()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.models import Recipe
from django.core import serializers


class Command(BaseCommand):
  help = 'Replace task or field in recipes with new value.'

  def add_arguments(self, parser):
    parser.add_argument(
        '--task',
        action='store',
        dest='task',
        default=None,
        help='Task to edit.',
    )

    parser.add_argument(
        '--field',
        action='store',
        dest='field',
        default=None,
        help='Task to edit.',
    )

    parser.add_argument(
        '--replacement',
        action='store',
        dest='replacement',
        default=None,
        help='Key to replace with.',
    )

    parser.add_argument(
        '--write',
        action='store_true',
        dest='write',
        default=False,
        help='Key to replace with.')

  def handle(self, *args, **kwargs):
    for recipe in (Recipe.objects.all()):

      print('---------------------------------------')
      print('Name:', recipe.name)
      print('Account:', recipe.account.email)

      changed = False
      tasks = json.loads(recipe.tasks)

      if kwargs['replacement']:
        for task in tasks:
          if kwargs['field']:
            if kwargs['field'] in task['values']:
              task['values'][kwargs['replacement']] = task['values'][
                  kwargs['field']]
              del task['values'][kwargs['field']]
              changed = True
          else:
            if task['tag'] == kwargs['task']:
              task['tag'] = kwargs['replacement']
              changed = True

      if changed:
        print('OLD TASK', recipe.tasks)
        print('NEW TASK', tasks)
        if kwargs['write']:
          recipe.set_values(tasks)
          recipe.save(update_fields=['tasks'])



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.models import Recipe


class Command(BaseCommand):
  help = 'Perform admin operations on recipes'

  def add_arguments(self, parser):
    parser.add_argument(
        '--recipe',
        action='store',
        dest='recipe',
        default=None,
        help='Run a specific recipe.',
    )

    parser.add_argument(
        '--activate',
        action='store_true',
        dest='activate',
        default=False,
        help='Activate specified recipe.',
    )

    parser.add_argument(
        '--deactivate',
        action='store_true',
        dest='deactivate',
        default=False,
        help='Deactivate specified recipe.',
    )

  def handle(self, *args, **kwargs):
    for recipe in (Recipe.objects.filter(pk=kwargs['recipe'])
                   if kwargs['recipe'] else Recipe.objects.filter(active=True)):
      try:
        if kwargs['activate'] == kwargs['deactivate']:
          raise Exception(
              'Specify one of the supported flags --activate or --deactivate')
        elif kwargs['activate']:
          recipe.activate()
        elif kwargs['deactivate']:
          recipe.deactivate()
        else:
          raise Exception('Unexpected exception')

        print('---------------------------------------')
        print('Name:', recipe.name)
        print('Account:', recipe.account.email)
        print('UUID:', recipe.uid())
        print('Reference:', recipe.reference)
        print('Active:', recipe.active)
        print('Week:', recipe.week)
        print('Hour:', recipe.hour)
        print('Timezone:', recipe.timezone)
        print('---------------------------------------')

      except (KeyboardInterrupt, SystemExit):
        raise

      except Exception as e:
        print('DEPLOY ERROR:', str(e))



import json
from urllib.request import Request, urlopen


def dynamite_write(room, key, token, text):
  url = 'https://chat.googleapis.com/v1/spaces/%s/messages?key=%s&token=%s' % (
      room, key, token)
  data = json.dumps({
      'sender': {
          'displayName':
              'StarThinker',
          'avatarUrl':
              'https://www.gstatic.com/images/icons/material/system/2x/chat_googblue_48dp.png'
      },
      'text': text
  })
  f = urlopen(
      urllib.request.Request(url, data, {
          'Content-Type': 'application/json',
          'Content-Length': len(data)
      }))
  response = f.read()
  f.close()
  return response



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.data import put_rows, get_rows
from starthinker.util.dbm import lineitem_read, lineitem_write
from starthinker.util.dbm.schema import LineItem_Read_Schema


@project.from_parameters
def lineitem():
  if project.verbose:
    print('LINEITEM')

  if 'read' in project.task:
    advertisers = []
    insertion_orders = []
    line_items = []

    if 'advertisers' in project.task['read']:
      advertisers = get_rows(project.task['auth'],
                             project.task['read']['advertisers'])

    elif 'insertion_orders' in project.task['read']:
      insertion_orders = get_rows(project.task['auth'],
                                  project.task['read']['insertion_orders'])

    elif 'line_items' in project.task['read']:
      line_items = list(
          get_rows(project.task['auth'], project.task['read']['line_items']))
      print('LI', line_items)

    rows = lineitem_read(project.task['auth'], advertisers, insertion_orders,
                         line_items)

    if rows:
      if 'bigquery' in project.task['read']['out']:
        project.task['read']['out']['bigquery']['schema'] = LineItem_Read_Schema
        project.task['read']['out']['bigquery']['skip_rows'] = 0

      put_rows(project.task['auth'], project.task['read']['out'], rows)

  elif 'write' in project.task:
    rows = get_rows(project.task['auth'], project.task['write'])
    lineitem_write(project.task['auth'], rows,
                   project.task['write'].get('dry_run', True))


if __name__ == '__main__':
  lineitem()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import argparse
import textwrap
from json import JSONDecodeError

from starthinker.util.project import get_project


def main():

  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Evaluate the validity of a json file. Helps in debugging recipes.
      Print the line and character position of any errors in the given json file.

      Example: python helper.py scripts/say_hello.json
  """))

  parser.add_argument('file', help='A JSON file.')
  args = parser.parse_args()

  try:
    project = get_project(args.file)
    print('JSON OK:', args.file)
  except JSONDecodeError as e:
    print(str(e))


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of placement groups."""

from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.campaign import CampaignDAO
from starthinker.task.traffic.feed import FieldMap


class PlacementGroupDAO(BaseDAO):
  """Placement group data access object.

  Inherits from BaseDAO and implements placement group specific logic for
  creating
  and
  updating placement group.
  """

  def __init__(self, auth, profile_id, is_admin):
    """Initializes PlacementGroupDAO with profile id and authentication scheme."""

    super(PlacementGroupDAO, self).__init__(auth, profile_id, is_admin)

    self.campaign_dao = CampaignDAO(auth, profile_id, is_admin)

    self._id_field = FieldMap.PLACEMENT_GROUP_ID
    self._search_field = FieldMap.PLACEMENT_GROUP_NAME
    self._list_name = 'placementGroups'
    self._entity = 'PLACEMENT_GROUP'

    self._parent_filter_name = 'campaignIds'
    self._parent_filter_field_name = FieldMap.CAMPAIGN_ID
    self._parent_dao = self.campaign_dao

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(PlacementGroupDAO, self)._api(iterate).placementGroups()

  def _process_update(self, item, feed_item):
    """Updates a placement group based on the values from the feed.

    Args:
      item: Object representing the placement group to be updated, this object
        is updated directly.
      feed_item: Feed item representing placement group values from the
        Bulkdozer feed.
    """
    campaign = self.campaign_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
    feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

    item['name'] = feed_item.get(FieldMap.PLACEMENT_GROUP_NAME, None)
    item['placementGroupType'] = feed_item.get(FieldMap.PLACEMENT_GROUP_TYPE,
                                               None)
    item['pricingSchedule']['startDate'] = feed_item.get(
        FieldMap.PLACEMENT_GROUP_START_DATE, None)
    item['pricingSchedule']['endDate'] = feed_item.get(
        FieldMap.PLACEMENT_GROUP_END_DATE, None)
    item['pricingSchedule']['pricingType'] = feed_item.get(
        FieldMap.PLACEMENT_GROUP_PRICING_TYPE, None)

  def _process_new(self, feed_item):
    """Creates a new placement group DCM object from a feed item representing a placement group from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the placement group from the Bulkdozer
        feed.

    Returns:
      A placement group object ready to be inserted in DCM through the API.

    """
    campaign = self.campaign_dao.get(feed_item, required=True)

    feed_item[FieldMap.CAMPAIGN_ID] = campaign['id']
    feed_item[FieldMap.CAMPAIGN_NAME] = campaign['name']

    return {
        'advertiserId':
            feed_item.get(FieldMap.ADVERTISER_ID, None),
        'campaignId':
            campaign['id'] if campaign else None,
        'siteId':
            feed_item.get(FieldMap.SITE_ID, None),
        'name':
            feed_item.get(FieldMap.PLACEMENT_GROUP_NAME, None),
        'placementGroupType':
            feed_item.get(FieldMap.PLACEMENT_GROUP_TYPE, None),
        'pricingSchedule': {
            'startDate':
                feed_item.get(FieldMap.PLACEMENT_GROUP_START_DATE, None),
            'endDate':
                feed_item.get(FieldMap.PLACEMENT_GROUP_END_DATE, None),
            'pricingType':
                feed_item.get(FieldMap.PLACEMENT_GROUP_PRICING_TYPE, None)
        }
    }



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

CM User Audit

Gives CM clients ability to see which profiles have access to which parts of an account. Loads CM user profile mappings using the API into BigQuery and connects to a DataStudio dashboard.

  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->CM_*</b> to be created.
  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->Barnacle_*</b> to be created, then copy and connect the following data sources.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/open/1a6K-XdPUzCYRXZp1ZcmeOUOURc9wn2Jj' target='_blank'>Barnacle Profile Advertiser Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1NEzrQWWnPjkD90iUwN-ASKbVBzoeBdoT' target='_blank'>Barnacle Profile Campaign Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1v_GRaitwPaHHKUzfJZYNBhzotvZ-bR7Y' target='_blank'>Barnacle Profile Site Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/14tWlh7yiqzxKJIppMFVOw2MoMtQV_ucE' target='_blank'>Barnacle Profiles Connections</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1mavjxvHSEPfJq5aW4FYgCXsBCE5rthZG' target='_blank'>Barnacle Report Delivery Profiles</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1Azk_Nul-auinf4NnDq8T9fDyiKkUWD7A' target='_blank'>Barnacle Roles Duplicates</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1ogoofpKtqkLwcW9qC_Ju_JvJdIajsjNI' target='_blank'>Barnacle Roles Not Used</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1xLgZPjOPDtmPyEqYMiMbWwMI8-WTslfj' target='_blank'>Barnacle Site Contacts Profiles</a> and connect.
  - If reports checked, copy <a href='https://datastudio.google.com/open/1-YGDiQPDnk0gD78_QOY5XdTXRlTrLeEq' target='_blank'>Barnacle Profile Report Map</a> and connect.
  - Copy <a href='https://datastudio.google.com/open/1gjxHm0jUlQUd0jMuxaOlmrl8gOX1kyKT' target='_blank'>Barnacle Report</a>.
  - When prompted choose the new data sources you just created.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_slug': '',  # Place where tables will be written in BigQuery.
  'auth_read': 'user',  # Credentials used for reading data.
  'recipe_project': '',  # Project where BigQuery dataset will be created.
  'auth_write': 'service',  # Credentials used for writing data.
  'accounts': [],  # Comma separated CM account ids.
  'reports': False,  # Include report audit, consumes significant API and data.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      1,
      3,
      8
    ]
  },
  'tasks': [
    {
      'dataset': {
        'description': 'The dataset will hold multiple tables, make sure it exists.',
        'hour': [
          1
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        }
      }
    },
    {
      'barnacle': {
        'description': 'Will create tables with format CM_* to hold each endpoint via a call to the API list function. Exclude reports for its own task.',
        'hour': [
          1
        ],
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'reports': {
          'field': {
            'name': 'reports',
            'kind': 'boolean',
            'order': 3,
            'default': False,
            'description': 'Include report audit, consumes significant API and data.'
          }
        },
        'accounts': {
          'single_cell': True,
          'values': {
            'field': {
              'name': 'accounts',
              'kind': 'integer_list',
              'order': 2,
              'default': [
              ],
              'description': 'Comma separated CM account ids.'
            }
          }
        },
        'out': {
          'auth': {
            'field': {
              'name': 'auth_write',
              'kind': 'authentication',
              'order': 1,
              'default': 'service',
              'description': 'Credentials used for writing data.'
            }
          },
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Google BigQuery dataset to create tables in.'
            }
          }
        }
      }
    },
    {
      'bigquery': {
        'hour': [
          8
        ],
        'description': 'Combine profile, account, subaccount, and roles into one view, used by other views in this workflow.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': " SELECT   P.profileId AS profileId,   P.accountId AS accountId,   P.subaccountId AS subaccountId,   P.name AS Profile_Name,   P.email AS Profile_Email,   REGEXP_EXTRACT(P.email, r'@(.+)') AS Profile_Domain,   P.userAccessType AS Profile_userAccessType,   P.active AS Profie_active,   P.traffickerType AS Profile_traffickerType,   P.comments AS Profile_comments,   P.userRoleId AS Profile_userRoleId,   R.role_name AS Role_role_name,   R.role_defaultUserRole AS Role_role_defaultUserRole,   R.permission_name AS Role_permission_name,   R.permission_availability AS Role_permission_availability,   A.name AS Account_name,   A.active AS Account_active,   A.description AS Account_description,   A.locale AS Account_locale,   S.name AS SubAccount_name FROM `[PARAMETER].[PARAMETER].CM_Profiles` AS P LEFT JOIN `[PARAMETER].[PARAMETER].CM_Roles` AS R   ON P.userRoleId=R.roleId LEFT JOIN `[PARAMETER].[PARAMETER].CM_Accounts` AS A   ON P.accountId=A.accountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_SubAccounts` AS S   ON P.accountId=S.accountId   AND P.subaccountId=S.subaccountId ; ",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profile_Role_Account_SubAccount_Map'
        }
      }
    },
    {
      'bigquery': {
        'description': 'Combine profiles and advertisers.',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   APRASM.*,   A.advertiserId AS advertiserId,   A.name AS Advertiser_name,   A.status AS Advertiser_status,   A.defaultEmail AS Advertiser_defaultEmail,   A.suspended AS Advertiser_suspended FROM `[PARAMETER].[PARAMETER].CM_Profile_Advertisers` As PA LEFT JOIN `[PARAMETER].[PARAMETER].Barnacle_Profile_Role_Account_SubAccount_Map` AS APRASM   ON PA.profileID=APRASM.profileID LEFT JOIN `[PARAMETER].[PARAMETER].CM_Advertisers` AS A   ON PA.advertiserId=A.advertiserId ; ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profile_Advertiser_Map'
        }
      }
    },
    {
      'bigquery': {
        'description': 'Profile to campaign mapping.',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   APRASM.*,   C.campaignId AS campaignId,   C.name AS Campaign_name,   C.archived AS Campaign_archived,   IF(C.startDate <= CURRENT_DATE() AND C.endDate >= CURRENT_DATE(), True, False) AS Campaign_running,   ROUND(TIMESTAMP_DIFF(CURRENT_TIMESTAMP(), C.lastModifiedInfo_time, DAY) / 7) AS Campaign_Modified_Weeks_Ago FROM `[PARAMETER].[PARAMETER].CM_Profile_Campaigns` As PC LEFT JOIN `[PARAMETER].[PARAMETER].Barnacle_Profile_Role_Account_SubAccount_Map` AS APRASM   ON PC.profileID=APRASM.profileID LEFT JOIN `[PARAMETER].[PARAMETER].CM_Campaigns` AS C   ON PC.campaignId=C.campaignId ; ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profile_Campaign_Map'
        }
      }
    },
    {
      'bigquery': {
        'description': 'The logic query for Deal Finder, transforms report into view used by datastudio.',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   APRASM.*,   R.reportId AS reportId,   R.name AS Report_name,   R.type AS Report_type,   R.format AS Report_format,   R.schedule_active AS Report_schedule_active,   R.schedule_repeats AS Report_schedule_repeats,   ROUND(TIMESTAMP_DIFF(CURRENT_TIMESTAMP(), R.lastModifiedTime, DAY) / 7) AS Report_Modified_Weeks_Ago,   DATE_DIFF(R.schedule_expirationDate, CURRENT_DATE(), MONTH) AS Report_Schedule_Weeks_To_Go FROM `[PARAMETER].[PARAMETER].CM_Reports` As R LEFT JOIN `[PARAMETER].[PARAMETER].Barnacle_Profile_Role_Account_SubAccount_Map` AS APRASM   ON R.profileID=APRASM.profileID ; ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profile_Report_Map'
        }
      }
    },
    {
      'bigquery': {
        'description': 'The logic query for Deal Finder, transforms report into view used by datastudio.',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   APRASM.*,   S.siteId AS siteId,   S.name AS Site_Name,   S.keyName AS Site_keyName,   S.approved AS Site_approved FROM `[PARAMETER].[PARAMETER].CM_Profile_Sites` As PS LEFT JOIN `[PARAMETER].[PARAMETER].Barnacle_Profile_Role_Account_SubAccount_Map` AS APRASM   ON PS.profileID=APRASM.profileID LEFT JOIN `[PARAMETER].[PARAMETER].CM_Sites` AS S   ON PS.siteId=S.siteId ; ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profile_Site_Map'
        }
      }
    },
    {
      'bigquery': {
        'description': 'The logic query for Deal Finder, transforms report into view used by datastudio.',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   APRASM.* FROM `[PARAMETER].[PARAMETER].Barnacle_Profile_Role_Account_SubAccount_Map` AS APRASM LEFT JOIN `[PARAMETER].[PARAMETER].CM_Profile_Advertisers` AS PA   ON APRASM.profileId=PA.profileId LEFT JOIN `[PARAMETER].[PARAMETER].CM_Profile_Campaigns` AS PC   ON APRASM.profileId=PC.profileId LEFT JOIN `[PARAMETER].[PARAMETER].CM_Profile_Sites` AS PS   ON APRASM.profileId=PS.profileId WHERE   PA.advertiserId IS NULL   AND PC.campaignId IS NULL   AND PS.siteId IS NULL  ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Profiles_Connections'
        }
      }
    },
    {
      'bigquery': {
        'description': '',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   RD.accountId AS accountId,   RD.subaccountId AS subaccountId,   RD.reportId AS reportId,   A.name AS Account_name,   A.active AS Account_active,   SA.name as SubAccount_name,   R.name as Report_name,   R.schedule_active AS Report_schedule_active,   RD.emailOwnerDeliveryType AS Delivery_emailOwnerDeliveryType,   RD.deliveryType AS Delivery_deliveryType,   RD.email AS Delivery_email,   RD.message AS Delivery_message,   IF(RD.email in (SELECT email from `[PARAMETER].[PARAMETER].CM_Profiles`), True, False) AS Profile_Match_Exists FROM `[PARAMETER].[PARAMETER].CM_Report_Deliveries` AS RD LEFT JOIN `[PARAMETER].[PARAMETER].CM_Accounts` AS A   ON RD.accountId=A.accountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_SubAccounts` AS SA   ON RD.subaccountId=SA.subaccountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_Reports` AS R   ON RD.reportId=R.reportId ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Report_Delivery_Profiles'
        }
      }
    },
    {
      'bigquery': {
        'description': '',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': ' SELECT   R.accountId AS accountId,   R.subaccountId AS subaccountId,   R.roleId AS roleId,   A.name AS Account_name,   A.active AS Account_active,   SA.name AS SubAccount_name,   R.role_name as Role_role_name,   R.role_defaultUserRole AS Role_role_defaultUserRole,   R.permission_name AS Role_permission_name,   R.permission_availability AS Role_permission_availability FROM `[PARAMETER].[PARAMETER].CM_Roles` AS R LEFT JOIN `[PARAMETER].[PARAMETER].CM_Accounts` AS A on R.accountId=A.accountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_SubAccounts` AS SA on R.subaccountId=SA.subaccountId WHERE roleId NOT IN (   SELECT roleId FROM `[PARAMETER].[PARAMETER].CM_Profile_Roles` ) ',
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Roles_Not_Used'
        }
      }
    },
    {
      'bigquery': {
        'description': '',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': " SELECT   SC.accountId AS accountId,   SC.subaccountId AS subaccountId,   SC.siteId AS siteId,   SC.contactId AS contactId,   A.name AS Account_name,   A.active AS Account_active,   SA.name as SubAccount_name,   S.name as Site_name,   S.approved AS Site_approved,   SC.email AS Site_Contact_email,   CONCAT(SC.firstName, ' ', sc.lastname) AS Site_Contact_Name,   SC.phone AS Site_Contact_phone,   SC.contactType AS Site_Contact_contactType,   IF(sc.email in (SELECT email from `[PARAMETER].[PARAMETER].CM_Profiles`), True, False) AS Profile_Match_Exists FROM `[PARAMETER].[PARAMETER].CM_Site_Contacts` AS SC LEFT JOIN `[PARAMETER].[PARAMETER].CM_Accounts` AS A   ON SC.accountId=A.accountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_SubAccounts` AS SA   ON SC.accountId=SA.accountId   AND SC.subaccountId=SA.subaccountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_Sites` AS S   ON SC.siteId=S.siteId ; ",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Site_Contacts_Profiles'
        }
      }
    },
    {
      'bigquery': {
        'description': '',
        'hour': [
          8
        ],
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'legacy': False,
          'query': " WITH  profile_counts AS (   SELECT userRoleId, COUNT(profileId) as profile_count   FROM `[PARAMETER].[PARAMETER].CM_Profiles`   GROUP BY 1 ),  permission_fingerprints AS (   SELECT     accountId,     subaccountId,     roleId,     role_name,     role_defaultUserRole,     SUM(profile_count) AS profile_count,     FARM_FINGERPRINT(       ARRAY_TO_STRING(       ARRAY_AGG(         DISTINCT permission_name ORDER BY permission_name ASC       ), ',', '-'     )   ) AS permissions_fingerprint   FROM     `[PARAMETER].[PARAMETER].CM_Roles` AS R   LEFT JOIN profile_counts AS P   ON R.roleId = P.userRoleId   GROUP BY     accountId,     subaccountId,     roleId,     role_name,     role_defaultUserRole )  SELECT   PFL.accountId AS accountId,   A.name AS Account_name,   A.active AS Account_active,   PFL.subaccountId AS subaccountId,   SA.name AS SubAccount_name,   PFL.roleId AS roleId,   PFL.role_name AS role_name,   PFL.role_defaultUserRole AS role_defaultUserRole,   COALESCE(PFL.profile_count, 0) AS profile_count,   PFR.roleId AS duplicate_roleId,   PFR.role_name AS duplicate_role_name,   PFR.role_defaultUserRole AS duplicate_role_defaultUserRole,   COALESCE(PFR.profile_count, 0) AS duplicate_profile_count FROM permission_fingerprints AS PFL LEFT JOIN `[PARAMETER].[PARAMETER].CM_Accounts` AS A on PFL.accountId=A.accountId LEFT JOIN `[PARAMETER].[PARAMETER].CM_SubAccounts` AS SA on PFL.subaccountId=SA.subaccountId LEFT JOIN permission_fingerprints AS PFR   ON PFL.permissions_fingerprint=PFR.permissions_fingerprint   AND PFL.accountId=PFR.accountId   AND COALESCE(PFL.subaccountId, 0)=COALESCE(PFR.subaccountId, 0) WHERE PFL.roleId != PFR.roleId ; ",
          'parameters': [
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            },
            {
              'field': {
                'name': 'recipe_project',
                'kind': 'string',
                'description': 'Project where BigQuery dataset will be created.'
              }
            },
            {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'description': 'Place where tables will be written in BigQuery.'
              }
            }
          ]
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          'view': 'Barnacle_Roles_Duplicates'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('barnacle', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import re
import json
import hashlib
from itertools import chain
from urllib.parse import quote_plus

from django import template
from django.conf import settings
from django.utils.html import mark_safe
from django.forms.widgets import CheckboxInput
from django.template.loader_tags import do_include

try:
  from django.utils.encoding import force_unicode
except:
  from django.utils.encoding import force_text as force_unicode

from starthinker_ui.recipe.forms_json import json_get_fields as json_get_fields_imported
from starthinker_ui.recipe.forms_fields import SwitchField
from starthinker_ui.recipe.scripts import Script

register = template.Library()


class SearchScriptsTag(template.Node):

  def __init__(self, filter):
    self.filter = filter

  def render(self, context):
    scripts = sorted(Script.get_scripts(ui=True), key=lambda x: x.get_name())

    if self.filter == 'SOME':
      scripts = [
          s for s in scripts if s.is_manual() == context.get('manual', False)
      ]

    if context.get('external', False):
      scripts = [s for s in scripts if s.get_open_source()]

    context['scripts'] = scripts
    context['agos'] = sorted(
        set(chain.from_iterable([s.get_released_ago() for s in scripts])))

    categories = {}
    for s in scripts:
      for f in s.get_from():
        categories.setdefault(f, [])
        categories[f].extend(s.get_to())
    for f in categories.keys():
      categories[f] = sorted(set(categories[f]))
    context['categories'] = sorted(categories.items())

    return ''


@register.tag('search_scripts')
def search_scripts(parser, token):
  tag, filter = token.split_contents()
  return SearchScriptsTag(filter)


@register.filter
def sort(value):
  return sorted(value)


@register.filter
def icon(value):
  return ''


@register.simple_tag
def request_solution(script):
  subject = 'Request For A %s StarThinker Solution' % script.get_name()
  body = ("Hi,\n\nI'd like to learn more about the %s solution.\n\nStarThinker "
          "Link: %s\n\nCan we set up some time to over it?\n\nThanks") % (
      script.get_name(), script.get_link())
  return mark_safe(
      'mailto:%s?subject=%s&body=%s' % (','.join(script.get_authors()).replace(
          ' ', ''), quote_plus(subject), quote_plus(body)))


@register.simple_tag
def mailto(emails, subject='', body=''):
  return mark_safe('mailto:%s?subject=%s&body=%s' % (','.join(emails).replace(
      ' ', ''), quote_plus(subject), quote_plus(body)))


@register.filter
def json_pretty(data):
  x = json.dumps(data, indent=4)
  return x


@register.filter
def is_checkbox(field):
  return isinstance(field.field.widget, CheckboxInput)


@register.filter
def is_switch(field):
  return isinstance(field.field, SwitchField)


@register.filter
def multiply(value, arg):
  return value * arg


@register.filter
def json_get_fields(value):
  return json_get_fields_imported(value)


@register.filter
def task_status_icon(event):

  if event == 'JOB_TIMEOUT':
    icon = 'alarm_off'
  elif event == 'JOB_ERROR':
    icon = 'error'
  elif event == 'JOB_END':
    icon = 'done_outline'
  elif event == 'JOB_START':
    icon = 'directions_walk'
  elif event == 'JOB_PENDING':
    icon = 'hourglass_empty'
  else:
    icon = 'hourglass_empty'

  return mark_safe(
      '<i class="small material-icons-outlined" style="vertical-align: middle;">%s</i>&nbsp;&nbsp;'
      % icon)


@register.filter
def calvin_id(name):
  return (5 *
          10**10) + int(hashlib.sha256(name.encode('utf-8')).hexdigest(),
                        16) % 10**9  # 5 + 9 digits


@register.filter
def email_to_ldap(email):
  try:
    return email.split('@')[0].lower()
  except:
    return ''


class GaplessNode(template.Node):

  def __init__(self, nodelist):
    self.nodelist = nodelist

  def render(self, context):
    return re.sub(r'\n\s*\n+', '\n',
                  force_unicode(self.nodelist.render(context).strip()))


@register.tag
def gapless(parser, token):
  nodelist = parser.parse(('endgapless',))
  parser.delete_first_token()
  return GaplessNode(nodelist)


class TryIncludeNode(template.Node):

  def __init__(self, parser, token):
    self.include_node = do_include(parser, token)

  def render(self, context):
    try:
      return self.include_node.render(context)
    except template.TemplateDoesNotExist:
      return ''


@register.tag('try_include')
def try_include(parser, token):
  return TryIncludeNode(parser, token)


@register.simple_tag
def google_analytics():
  if settings.GOOGLE_ANALYTICS:
    return mark_safe("""<!-- Global site tag (gtag.js) - Google Analytics -->
    <script async src="https://www.googletagmanager.com/gtag/js?id=%s"></script>
    <script>
      window.dataLayer = window.dataLayer || [];
      function gtag(){dataLayer.push(arguments);}
      gtag('js', new Date());
      gtag('config', '%s');
    </script>""" % (settings.GOOGLE_ANALYTICS, settings.GOOGLE_ANALYTICS))
  else:
    return ''



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets

import re

from googleapiclient.errors import HttpError

from starthinker.util.project import project
from starthinker.util.google_api import API_Sheets
from starthinker.util.drive import file_delete
from starthinker.util.drive import file_find


def sheets_id(auth, url_or_name):
  # check if URL given, convert to ID "https://docs.google.com/spreadsheets/d/1uN9tnb-DZ9zZflZsoW4_34sf34tw3ff/edit#gid=4715"
  if url_or_name.startswith('https://docs.google.com/spreadsheets/d/'):
    m = re.search(
        '^(?:https:\/\/docs.google.com\/spreadsheets\/d\/)?([a-zA-Z0-9-_]+)(?:\/.*)?$',
        url_or_name)
    if m:
      return m.group(1)

  # check if name given convert to ID "Some Document"
  else:
    sheet = file_find(auth, url_or_name)
    if sheet:
      return sheet['id']

      # check if just ID given, "1uN9tnb-DZ9zZflZsoW4_34sf34tw3ff"
    else:
      m = re.search('^([a-zA-Z0-9-_]+)$', url_or_name)
      if m:
        return m.group(1)

  # probably a mangled id or name does not exist
  if project.verbose:
    print('SHEET DOES NOT EXIST', url_or_name)
  return None


def sheets_url(auth, url_or_name):
  sheet_id = sheets_id(auth, url_or_name)
  return 'https://docs.google.com/spreadsheets/d/%s/' % sheet_id


def sheets_tab_range(sheet_tab, sheet_range):
  if sheet_range:
    return '%s!%s' % (sheet_tab, sheet_range)
  else:
    return sheet_tab


def sheets_get(auth, sheet_url_or_name):
  sheet_id = sheets_id(auth, sheet_url_or_name)
  if sheet_id:
    return API_Sheets(auth).spreadsheets().get(spreadsheetId=sheet_id).execute()
  else:
    return None


def sheets_tab_id(auth, sheet_url_or_name, sheet_tab):
  sheet_id = None
  tab_id = None
  spreadsheet = sheets_get(auth, sheet_url_or_name)
  if spreadsheet:
    sheet_id = spreadsheet['spreadsheetId']
    for tab in spreadsheet.get('sheets', []):
      if tab['properties']['title'] == sheet_tab:
        tab_id = tab['properties']['sheetId']
        break
  return sheet_id, tab_id


def sheets_read(auth, sheet_url_or_name, sheet_tab, sheet_range='', retries=10):
  if project.verbose:
    print('SHEETS READ', sheet_url_or_name, sheet_tab, sheet_range)
  sheet_id = sheets_id(auth, sheet_url_or_name)
  if sheet_id is None:
    raise (OSError('Sheet does not exist: %s' % sheet_url_or_name))
  else:
    return API_Sheets(auth).spreadsheets().values().get(
      spreadsheetId=sheet_id,
      range=sheets_tab_range(sheet_tab, sheet_range)
    ).execute().get('values')


# TIP: Specify sheet_range as 'Tab!A1' coordinate, the API will figure out length and height based on data
def sheets_write(auth,
                 sheet_url_or_name,
                 sheet_tab,
                 sheet_range,
                 data,
                 append=False,
                 valueInputOption='RAW'):
  if project.verbose:
    print('SHEETS WRITE', sheet_url_or_name, sheet_tab, sheet_range)
  sheet_id = sheets_id(auth, sheet_url_or_name)
  range = sheets_tab_range(sheet_tab, sheet_range)
  body = {'values': list(data)}

  if append:
    API_Sheets(auth).spreadsheets().values().append(
      spreadsheetId=sheet_id,
      range=range,
      body=body,
      valueInputOption=valueInputOption,
      insertDataOption='OVERWRITE'
    ).execute()
  else:
    API_Sheets(auth).spreadsheets().values().update(
      spreadsheetId=sheet_id,
      range=range,
      body=body,
      valueInputOption=valueInputOption
    ).execute()


def sheets_clear(auth, sheet_url_or_name, sheet_tab, sheet_range):
  if project.verbose:
    print('SHEETS CLEAR', sheet_url_or_name, sheet_tab, sheet_range)
  sheet_id = sheets_id(auth, sheet_url_or_name)
  API_Sheets(auth).spreadsheets().values().clear(
    spreadsheetId=sheet_id,
    range=sheets_tab_range(sheet_tab, sheet_range),
    body={}
  ).execute()


def sheets_tab_copy(auth,
                    from_sheet_url_or_name,
                    from_sheet_tab,
                    to_sheet_url_or_name,
                    to_sheet_tab,
                    overwrite=False):
  if project.verbose:
    print('SHEETS COPY', from_sheet_url_or_name, from_sheet_tab,
          to_sheet_url_or_name, to_sheet_tab)

  # convert human readable to ids
  from_sheet_id, from_tab_id = sheets_tab_id(auth, from_sheet_url_or_name,
                                             from_sheet_tab)
  to_sheet_id, to_tab_id = sheets_tab_id(auth, to_sheet_url_or_name,
                                         to_sheet_tab)

  # overwrite only if does not exist
  if overwrite or to_tab_id is None:

    # copy tab between sheets, the name changes to be "Copy of [from_sheet_tab]"
    copy_sheet = API_Sheets(auth).spreadsheets().sheets().copyTo(
      spreadsheetId=from_sheet_id,
      sheetId=from_tab_id,
      body={
        'destinationSpreadsheetId': to_sheet_id,
      }
    ).execute()

    body = {'requests': []}

    # if destination tab exists, delete it
    if to_tab_id:
      body['requests'].append({'deleteSheet': {'sheetId': to_tab_id}})

    # change the copy name to the designated name, remove "Copy of "
    body['requests'].append({
        'updateSheetProperties': {
            'properties': {
                'sheetId': copy_sheet['sheetId'],
                'title': to_sheet_tab
            },
            'fields': 'title'
        }
    })

    API_Sheets(auth).spreadsheets().batchUpdate(
      spreadsheetId=to_sheet_id,
      body=body
    ).execute()


def sheets_batch_update(auth, sheet_url_or_name, data):
  sheet_id = sheets_id(auth, sheet_url_or_name)
  API_Sheets(auth).spreadsheets().batchUpdate(
    spreadsheetId=sheet_id,
    body=data
  ).execute()


def sheets_values_batch_update(auth, sheet_url_or_name, data):
  sheet_id = sheets_id(auth, sheet_url_or_name)
  API_Sheets(auth).spreadsheets().values().batchUpdate(
    spreadsheetId=sheet_id,
    body=data
  ).execute()


def sheets_tab_create(auth, sheet_url_or_name, sheet_tab):
  sheet_id, tab_id = sheets_tab_id(auth, sheet_url_or_name, sheet_tab)
  if tab_id is None:
    sheets_batch_update(
        auth, sheet_url_or_name,
        {'requests': [{
            'addSheet': {
                'properties': {
                    'title': sheet_tab
                }
            }
        }]})


def sheets_tab_delete(auth, sheet_url_or_name, sheet_tab):
  if project.verbose:
    print('SHEETS DELETE', sheet_url_or_name, sheet_tab)

  spreadsheet = sheets_get(auth, sheet_url_or_name)
  if spreadsheet:
    if len(
        spreadsheet['sheets']
    ) == 1 and spreadsheet['sheets'][0]['properties']['title'] == sheet_tab:
      file_delete(auth, spreadsheet['properties']['title'], parent=None)
    else:
      sheet_id, tab_id = sheets_tab_id(auth, sheet_url_or_name, sheet_tab)
      # add check to see if only tab, then delete whole sheet
      if tab_id is not None:
        sheets_batch_update(
            auth, sheet_url_or_name,
            {'requests': [{
                'deleteSheet': {
                    'sheetId': tab_id,
                }
            }]})


def sheets_tab_rename(auth, sheet_url_or_name, old_sheet_tab, new_sheet_tab):
  sheet_id, tab_id = sheets_tab_id(auth, sheet_url_or_name, old_sheet_tab)
  if tab_id is not None:
    sheets_batch_update(
        auth, sheet_url_or_name, {
            'requests': [{
                'updateSheetProperties': {
                    'properties': {
                        'sheetId': tab_id,
                        'title': new_sheet_tab
                    },
                    'fields': 'title'
                }
            }]
        })


def sheets_create(auth,
                  sheet_name,
                  sheet_tab,
                  template_sheet=None,
                  template_tab=None):
  """ Checks if sheet with name already exists ( outside of trash ) and

  if not, creates the sheet. Both sheet and tab must be provided or both must be
  omitted to create
  a blank sheet and tab.

  Args:
    * auth: (string) Either user or service.
    * sheet_name: (string) name of sheet to create, used as key to check if it
      exists in the future.
    * sheet_tab: (string) name of the tab to create.
    * template_sheet: (string) optional sheet to copy tempalte from.
    * template_tab: (string) optional tab to copy template from.
    * parent: (string) the Google Drive to upload the file to.

  Returns:
    * JSON specification of the file created or existing.

  """

  created = False

  # check if sheet and tab exist
  sheet_id, tab_id = sheets_tab_id(auth, sheet_name, sheet_tab)

  # if no sheet create it and the tab
  if sheet_id is None:
    if project.verbose:
      print('SHEET CREATE', sheet_name, sheet_tab)
    body = {
        'properties': {
            'title': sheet_name,
        },
        'sheets': [{
            'properties': {
                'title': sheet_tab,
            }
        }]
    }
    spreadsheet = API_Sheets(auth).spreadsheets().create(body=body).execute()
    sheet_id = spreadsheet['spreadsheetId']
    tab_id = spreadsheet['sheets'][0]['properties']['title']
    created = True

  # if creating tab from template
  if (created or tab_id is None) and template_sheet and template_tab:
    if project.verbose:
      print('SHEET TAB COPY', sheet_tab)
    sheets_tab_copy(auth, template_sheet, template_tab, sheet_id, sheet_tab,
                    True)

  # if creating a blank tab
  elif tab_id is None:
    if project.verbose:
      print('SHEET TAB CREATE', sheet_name, sheet_tab)
    sheets_tab_create(auth, sheet_name, sheet_tab)

  # if sheet and tab already exist
  else:
    if project.verbose:
      print('SHEET EXISTS', sheet_name, sheet_tab)

  return sheet_id, tab_id, created



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Entity Read Files

Import public and private <a href='https://developers.google.com/bid-manager/guides/entity-read/format-v2' target='_blank'>Entity Read Files</a> into a BigQuery dataset.<br/>CAUTION: PARTNER ONLY, ADVERTISER FILTER IS NOT APPLIED.

  - Entity Read Files ONLY work at the partner level.
  - Advertiser filter is NOT APPLIED.
  - Specify one or more partners to be moved into the dataset.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_write': 'service',  # Credentials used for writing data.
  'auth_read': 'user',  # Credentials used for reading data.
  'partners': '[]',  # Comma sparated list of DV360 partners.
  'dataset': '',  # BigQuery dataset to write tables for each entity.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      5
    ]
  },
  'tasks': [
    {
      'dataset': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 3,
            'default': '',
            'description': 'BigQuery dataset to write tables for each entity.'
          }
        }
      }
    },
    {
      'entity': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'prefix': 'Entity',
        'entities': [
          'Campaign',
          'LineItem',
          'Creative',
          'UserList',
          'Partner',
          'Advertiser',
          'InsertionOrder',
          'Pixel',
          'InventorySource',
          'CustomAffinity',
          'UniversalChannel',
          'UniversalSite',
          'SupportedExchange',
          'DataPartner',
          'GeoLocation',
          'Language',
          'DeviceCriteria',
          'Browser',
          'Isp'
        ],
        'partners': {
          'single_cell': True,
          'values': {
            'field': {
              'name': 'partners',
              'kind': 'integer_list',
              'order': 1,
              'default': '[]',
              'description': 'Comma sparated list of DV360 partners.'
            }
          }
        },
        'out': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Credentials used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 3,
                'default': '',
                'description': 'BigQuery dataset to write tables for each entity.'
              }
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('entity', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()
